/////////////////////////////////////////////////////////////////////////////
// Name:       
// Purpose:  
// Author:      Cesar Mauri Loba (cesar at crea-si dot com)
// Modified by: 
// Created:     
// Copyright:   (C) 2010 Cesar Mauri Loba - CREA Software Systems
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////
#include "exceptions.h"
#include "cameracontroller.h"
#include "camselect.h"
#include "camwindow2.h"

#include <wx/config.h>

#if defined(WIN32)
	#include "crvcamera_wdm.h"
	#define CAM_CLASS CCameraWDM
#else 
	#include "crvcamera_cv.h"
	#define CAM_CLASS CCameraCV
#endif

DEFINE_EVENT_TYPE( wxEVT_CAMCTRL_CAM_STARTED )
DEFINE_EVENT_TYPE( wxEVT_CAMCTRL_FIRSTPROCESSOR )
DEFINE_EVENT_TYPE( wxEVT_CAMCTRL_LASTPROCESSOR )
//DEFINE_EVENT_TYPE( wxEVT_CAMCTRL_LASTDRAWER )
DEFINE_EVENT_TYPE( wxEVT_CAMCTRL_ERROR )
DEFINE_EVENT_TYPE( wxEVT_CAMCTRL_FATAL_ERROR )

BEGIN_EVENT_TABLE(CCameraController, wxEvtHandler)    
	EVT_MY_THREAD(CCameraController::OnThreadEvent)
END_EVENT_TABLE()

CCameraController* CCameraController::g_single= NULL;
	
CCameraController* CCameraController::getInstance(wxConfigBase* configBase, wxEvtHandler* evtDest)
{
	if (g_single== NULL)
	{
		assert (configBase);
		g_single= new CCameraController(configBase, evtDest);
	}
	return g_single;
}

void CCameraController::destroyInstance()
{
	if (g_single) {
		delete g_single;
		g_single= NULL;
	}
}

// Constructor. First parameter sets  configuration object to retrieve/store default
// camera, second parameter sets destination for capture thread events	
CCameraController::CCameraController(wxConfigBase* configBase, wxEvtHandler* evtDest)
: m_captureThread(NULL)
, m_pCamera(NULL)
, m_visibleNormROI(1.0f, 1.0f, 0.0f, 0.0f, false)
, m_evtDest(evtDest)
, m_configBase(configBase)
, m_wantToSelectCamera(false)
{
	assert (m_configBase);

	m_configPath= m_configBase->GetPath();
}

/*
// Constructor with a fixed camera
CCameraController::CCameraController(CCamera* camera, wxEvtHandler* evtDest)
: m_captureThread(NULL), m_pCamera(camera), m_evtDest(evtDest), m_configBase(NULL)
{
	assert (m_pCamera);
}
*/

CCameraController::~CCameraController(void)
{
	DestroyCaptureThread();
//	if (m_pCamera) delete m_pCamera;	
}

#define CAM_CONFIG_KEY _T("defaultCamera")

void CCameraController::InitCamera ()
{
	if (m_pCamera!= NULL) return;

	wxString prevPath (m_configBase->GetPath());
	//wxString camName;

	m_configBase->SetPath (m_configPath);

	wxConfig::Get()->Read (CAM_CONFIG_KEY, &m_camName);
	int camId= -1;
	if (!m_wantToSelectCamera) camId= CCamSelect::GetCamIdByName (m_camName);
	if (camId== -1) {
		try {
			camId= CCamSelect::SelectCameraInteractive (m_camName, _("Select default camera"), m_wantToSelectCamera);
			m_wantToSelectCamera= false;
		}
		catch(...)
		{
			m_configBase->SetPath (prevPath);
			wxConfig::Get()->Write (CAM_CONFIG_KEY, m_camName);
			throw;
		}
		assert (camId!= -1);
		wxConfig::Get()->Write (CAM_CONFIG_KEY, m_camName);
    }
	m_configBase->SetPath (prevPath);

	m_pCamera= new CAM_CLASS(camId);
	if (m_pCamera== NULL)
		throw CameraException (_("Can't initialize camera."));
//	if (!m_pCamera->Open ())
//		throw CameraException (_("Unexpected error opening the camera."));
	m_pCamera->SetHorizontalFlip(true);
}

void CCameraController::DestroyCaptureThread()
{
	if (m_captureThread) {
		// Set m_captureThread NULL to avoid reentrant calls
		CCaptureThread2 *ct= m_captureThread;
		m_captureThread= NULL;

		ct->Finish ();
		ct->Wait ();
		delete ct; 

		if (m_pCamera) {
			//m_pCamera->Close();
			delete m_pCamera;
			m_pCamera= NULL;
		}
	}
}

void CCameraController::CheckStartThread()
{	
	if (m_captureThread== NULL) {
		InitCamera ();
		m_captureThread= new CCaptureThread2(m_pCamera, wxTHREAD_JOINABLE, this);
		m_captureThread->Create();
		m_captureThread->Run();		
	}
}

void CCameraController::CheckStopThread()
{
	if (m_captureThread)
		if (m_captureThread->GetRegisteredImageProcessors()== 0 && m_captureThread->GetRegisteredImageDrawers()== 0)
			DestroyCaptureThread();
}

void CCameraController::RegisterImageProcessor (CProcessImage2* iProcessor)
{
	assert (iProcessor);

	bool firstProcessor= false;
	{
		wxCriticalSectionLocker mutex(m_mutex);
		CheckStartThread();
		if (m_captureThread->RegisterImageProcessor (iProcessor)== 1) firstProcessor= true;		
	}
	if (firstProcessor && m_evtDest) {
		wxCommandEvent event (wxEVT_CAMCTRL_FIRSTPROCESSOR);
		wxPostEvent (m_evtDest, event);
	}
}

void CCameraController::UnregisterImageProcessor (CProcessImage2* iProcessor)
{
	assert (iProcessor);

	bool sendEvent= false;
	{
		wxCriticalSectionLocker mutex(m_mutex);

		// consider removing assert when external plug-ins exists
		assert (m_captureThread);

		if (m_captureThread) {
			if (m_captureThread->UnregisterImageProcessor (iProcessor)== 0) sendEvent= true;
			CheckStopThread();
		}
	}
	if (sendEvent && m_evtDest) {
		wxCommandEvent event (wxEVT_CAMCTRL_LASTPROCESSOR);
		wxPostEvent (m_evtDest, event);
	}
}

void CCameraController::RegisterImageDrawer (CCamWindow2* iDrawer)
{
	assert (iDrawer);

	wxCriticalSectionLocker mutex(m_mutex);
	CheckStartThread();
	/*
	if (m_captureThread== NULL) {
		InitCamera ();
		m_captureThread= new CCaptureThread2(m_pCamera, wxTHREAD_JOINABLE, this);
		m_captureThread->Create();
		m_captureThread->Run();		
	}*/
	iDrawer->SetVisibleNormROI(&m_visibleNormROI);
	m_captureThread->RegisterImageDrawer (iDrawer);
}

void CCameraController::UnregisterImageDrawer (CCamWindow2* iDrawer)
{	
	assert (iDrawer);

	wxCriticalSectionLocker mutex(m_mutex);

	// consider removing assert when external plug-ins exists
	assert (m_captureThread);

	if (m_captureThread) {
		m_captureThread->UnregisterImageDrawer (iDrawer);
		iDrawer->SetVisibleNormROI(NULL);
		CheckStopThread();
	}
}

void CCameraController::RegisterChildROI2 (CVisibleNormROI2* nr2)
{
	m_visibleNormROI.RegisterChildROI (nr2);
}

void CCameraController::UnregisterChildROI2 (CVisibleNormROI2* nr2)
{
    m_visibleNormROI.UnregisterChildROI (nr2);
}

int CCameraController::GetImageWidth () const
{
	//wxCriticalSectionLocker mutex(m_mutex);
	if (m_pCamera)
		return m_pCamera->GetRealWidth();

	return 0;
}

int CCameraController::GetImageHeight () const
{
	//wxCriticalSectionLocker mutex(m_mutex);
	if (m_pCamera)
		return m_pCamera->GetRealHeight();

	return 0;
}

float CCameraController::GetFrameRate () const
{
	//wxCriticalSectionLocker mutex(m_mutex);/
	if (m_pCamera)
		return m_pCamera->GetRealFrameRate();

	return 0;
}

const wxString& CCameraController::GetCameraName() const
{
	return m_camName;
}

void CCameraController::ShowCameraSettings()
{
	wxCriticalSectionLocker mutex(m_mutex);
	if (m_pCamera)
		m_pCamera->ShowSettingsDialog();
}

void CCameraController::OnThreadEvent ( myThreadEvent &event )
{
	if (event.GetFlags() & myThreadEvent::STARTED) {
		if (m_evtDest) {
			wxCommandEvent event2 (wxEVT_CAMCTRL_CAM_STARTED);
			wxPostEvent (m_evtDest, event2);
		}
	}
	if (event.GetFlags() & myThreadEvent::ENDED) CheckStopThread();
	if (event.GetFlags() & myThreadEvent::FATAL_ERROR) {
		if (m_evtDest) {
			//wxPostEvent (m_evtDest, event);
			wxCommandEvent event2 (wxEVT_CAMCTRL_FATAL_ERROR);
			wxPostEvent (m_evtDest, event2);
		}
	}
	if (event.GetFlags() & myThreadEvent::NON_FATAL_ERROR) {
		if (m_evtDest) {
			//wxPostEvent (m_evtDest, event);
			wxCommandEvent event2 (wxEVT_CAMCTRL_ERROR);
			wxPostEvent (m_evtDest, event2);
		}
		if (m_captureThread->GetRegisteredImageProcessors()== 0) {
			if (m_evtDest) {
				wxCommandEvent event2 (wxEVT_CAMCTRL_LASTPROCESSOR);
				wxPostEvent (m_evtDest, event2);
			}
		}
		CheckStopThread();
	}

	event.Skip(true);
}

////////////////////////////////////////////////

CCameraProcessor::CCameraProcessor (CProcessImage2& iProcessor)
: m_started(false)
{
	m_iProcessor= &iProcessor;
}

CCameraProcessor::~CCameraProcessor ()
{
	Unregister();
}

void CCameraProcessor::Register()
{
	assert (!m_started);
	if (m_started) return;

	CCameraController::getInstance()->RegisterImageProcessor (m_iProcessor);
	m_started= true;
}

void CCameraProcessor::Unregister()
{
	if (m_started)
	{
		m_started= false;
		CCameraController::getInstance()->UnregisterImageProcessor (m_iProcessor);
	}
}

////////////////////////////////////////////////

CCameraDrawer::CCameraDrawer (CCamWindow2& iDrawer)
: m_started(false)
{
	m_iDrawer= &iDrawer;
}

CCameraDrawer::~CCameraDrawer ()
{
	Unregister();
}

void CCameraDrawer::Register()
{
	assert (!m_started);
	if (m_started) return;

	CCameraController::getInstance()->RegisterImageDrawer (m_iDrawer);
	m_started= true;
}

void CCameraDrawer::Unregister()
{
	if (m_started)
	{
		m_started= false;
		CCameraController::getInstance()->UnregisterImageDrawer (m_iDrawer);
	}
}

////////////////////////////////////////////////

CVisibleNormROI2Instance::CVisibleNormROI2Instance(void)
{
	m_instance= new CVisibleNormROI2();
	CCameraController::getInstance()->RegisterChildROI2 (m_instance);
}

CVisibleNormROI2Instance::CVisibleNormROI2Instance(float width, float height, float x, float y, bool visible)
{
	m_instance= new CVisibleNormROI2(width, height, x, y, visible);
	CCameraController::getInstance()->RegisterChildROI2 (m_instance);
}

CVisibleNormROI2Instance::~CVisibleNormROI2Instance(void)
{
	if (m_instance) {
		delete m_instance;
		m_instance= NULL;
	}
}
